from __future__ import absolute_import, unicode_literals

from datetime import datetime
import types
from django.utils import timezone
import pytest
from souvenirs.utils import (adjust_to_calendar_month,
                             adjust_to_subscription_start,
                             iter_days, iter_months, iter_quarters, iter_years,
                             next_month, nearest_dom)


def test_adjust_to_calendar_month():
    dt = timezone.make_aware(datetime(2017, 3, 27, 10, 36))
    assert (adjust_to_calendar_month(dt) ==
            timezone.make_aware(datetime(2017, 3, 1, 0, 0, 0, 0)))


def test_adjust_to_subscription_start():
    ss = timezone.make_aware(datetime(2017, 1, 31, 23, 30))

    # if dt matches ss, go with it
    dt = timezone.make_aware(datetime(2017, 3, 31, 23, 30))
    assert adjust_to_subscription_start(dt, ss) == dt

    # if dt is after ss, simple adjust back to ss
    dt = timezone.make_aware(datetime(2017, 3, 31, 23, 31))
    assert (adjust_to_subscription_start(dt, ss) ==
            timezone.make_aware(datetime(2017, 3, 31, 23, 30)))

    # if dt is before ss, adjust back to previous month
    dt = timezone.make_aware(datetime(2017, 3, 27, 10, 41))
    assert (adjust_to_subscription_start(dt, ss) ==
            timezone.make_aware(datetime(2017, 2, 28, 23, 30)))

    # interesting scenario of dt matching adjusted ss
    ss = timezone.make_aware(datetime(2016, 12, 31, 12, 0))
    dt = timezone.make_aware(datetime(2017, 2, 28, 12, 0))
    assert adjust_to_subscription_start(dt, ss) == dt

    # interesting scenario of dt after adjusted ss
    ss = timezone.make_aware(datetime(2016, 12, 31, 12, 0))
    dt = timezone.make_aware(datetime(2017, 2, 28, 13, 0))
    assert (adjust_to_subscription_start(dt, ss) ==
            timezone.make_aware(datetime(2017, 2, 28, 12, 0)))


def test_iter_days():
    start = timezone.make_aware(datetime(2017, 3, 29, 1, 2, 3))
    end = timezone.make_aware(datetime(2017, 4, 7, 11, 22, 33))

    days = iter_days(start, end)

    assert type(days) is types.GeneratorType

    starts = [
        datetime.combine(datetime(year, month, day).date(), start.timetz())
        for year, month, day in [
                (2017, 3, 29),
                (2017, 3, 30),
                (2017, 3, 31),
                (2017, 4, 1),
                (2017, 4, 2),
                (2017, 4, 3),
                (2017, 4, 4),
                (2017, 4, 5),
                (2017, 4, 6),
                (2017, 4, 7),
        ]
    ]

    ends = starts[1:] + [end]

    assert list(days) == list(zip(starts, ends))


def test_iter_months():
    start = timezone.make_aware(datetime(2015, 12, 31, 1, 2, 3))
    end = timezone.make_aware(datetime(2018, 6, 10, 11, 22, 33))

    months = iter_months(start, end)

    assert type(months) is types.GeneratorType

    starts = [
        datetime.combine(datetime(year, month, day).date(), start.timetz())
        for year, month, day in [
                (2015, 12, 31),
                (2016, 1, 31),
                (2016, 2, 29),  # leap!
                (2016, 3, 31),
                (2016, 4, 30),
                (2016, 5, 31),
                (2016, 6, 30),
                (2016, 7, 31),
                (2016, 8, 31),
                (2016, 9, 30),
                (2016, 10, 31),
                (2016, 11, 30),
                (2016, 12, 31),
                (2017, 1, 31),
                (2017, 2, 28),
                (2017, 3, 31),
                (2017, 4, 30),
                (2017, 5, 31),
                (2017, 6, 30),
                (2017, 7, 31),
                (2017, 8, 31),
                (2017, 9, 30),
                (2017, 10, 31),
                (2017, 11, 30),
                (2017, 12, 31),
                (2018, 1, 31),
                (2018, 2, 28),
                (2018, 3, 31),
                (2018, 4, 30),
                (2018, 5, 31),
        ]
    ]

    ends = starts[1:] + [end]

    assert list(months) == list(zip(starts, ends))


def test_iter_quarters():
    start = timezone.make_aware(datetime(2015, 11, 30, 1, 2, 3))
    end = timezone.make_aware(datetime(2017, 2, 28, 11, 22, 33))

    quarters = iter_quarters(start, end)

    assert type(quarters) is types.GeneratorType

    starts = [
        datetime.combine(datetime(year, month, day).date(), start.timetz())
        for year, month, day in [
                (2015, 11, 30),
                (2016, 2, 29),  # leap!
                (2016, 5, 30),
                (2016, 8, 30),
                (2016, 11, 30),
                (2017, 2, 28),
        ]
    ]

    ends = starts[1:] + [end]

    assert list(quarters) == list(zip(starts, ends))


def test_iter_years():
    start = timezone.make_aware(datetime(2016, 2, 29, 1, 2, 3))
    end = timezone.make_aware(datetime(2019, 2, 28, 11, 22, 33))

    years = iter_years(start, end)

    assert type(years) is types.GeneratorType

    starts = [
        datetime.combine(datetime(year, month, day).date(), start.timetz())
        for year, month, day in [
                (2016, 2, 29),  # leap!
                (2017, 2, 28),
                (2018, 2, 28),
                (2019, 2, 28),
        ]
    ]

    ends = starts[1:] + [end]

    assert list(years) == list(zip(starts, ends))


def test_next_month():
    dt = timezone.make_aware(datetime(2017, 3, 30, 11, 5))

    assert next_month(dt) == timezone.make_aware(datetime(2017, 4, 30, 11, 5))
    assert next_month(dt, delta=2) == timezone.make_aware(datetime(2017, 5, 30, 11, 5))
    assert next_month(dt, delta=12) == timezone.make_aware(datetime(2018, 3, 30, 11, 5))
    assert next_month(dt, delta=-1) == timezone.make_aware(datetime(2017, 2, 28, 11, 5))
    assert next_month(dt, delta=-12) == timezone.make_aware(datetime(2016, 3, 30, 11, 5))

    assert (next_month(dt, preferred_dom=31) ==
            timezone.make_aware(datetime(2017, 4, 30, 11, 5)))
    assert (next_month(dt, preferred_dom=31, delta=2)
            == timezone.make_aware(datetime(2017, 5, 31, 11, 5)))
    assert (next_month(dt, preferred_dom=31, delta=12)
            == timezone.make_aware(datetime(2018, 3, 31, 11, 5)))
    assert (next_month(dt, preferred_dom=31, delta=-1)
            == timezone.make_aware(datetime(2017, 2, 28, 11, 5)))
    assert (next_month(dt, preferred_dom=31, delta=-12)
            == timezone.make_aware(datetime(2016, 3, 31, 11, 5)))

    with pytest.raises(ValueError):
        next_month(dt, delta=0)


def test_nearest_dom():
    assert nearest_dom(2017, 1, 1) == 1
    assert nearest_dom(2017, 1, 31) == 31
    assert nearest_dom(2017, 2, 27) == 27
    assert nearest_dom(2017, 2, 28) == 28
    assert nearest_dom(2017, 2, 29) == 28
    assert nearest_dom(2017, 2, 30) == 28
    assert nearest_dom(2017, 2, 31) == 28




from __future__ import absolute_import, unicode_literals

from souvenirs.control import souvenez


class SouvenirsMiddleware(object):

    def process_request(self, request):
        if request.user.is_authenticated():
            souvenez(request.user)



from __future__ import absolute_import, unicode_literals

import calendar
from datetime import datetime, timedelta


def adjust_to_calendar_month(dt):
    """
    Return dt normalized to the start of the calendar month (day reset to 1,
    time reset to 00:00:00 local time).
    """
    return datetime(dt.year, dt.month, 1, tzinfo=dt.tzinfo)


def adjust_to_subscription_start(dt, ss):
    """
    Return dt adjusted back to the day, time and tzinfo matching ss.
    """
    new_dt = datetime.combine(
        dt.date().replace(day=nearest_dom(dt.year, dt.month, ss.day)),
        ss.timetz()
    )
    if new_dt > dt:
        new_dt = next_month(new_dt, preferred_dom=ss.day, delta=-1)
    return new_dt


def iter_days(start, end):
    """
    Generate a sequence of tuples representing the span of a day
    (day_start, day_end) where starts are inclusive and ends are exclusive (the
    end of one day is the same as the start of the next day).
    """
    while start < end:
        next_start = min(start + timedelta(days=1), end)
        yield start, next_start
        start = next_start


def iter_months(start, end):
    """
    Generate a sequence of tuples representing the span of a month
    (month_start, month_end) where starts are inclusive and ends are exclusive
    (the end of one month is the same as the start of the next month).
    """
    preferred_dom = start.day
    while start < end:
        next_start = min(next_month(start, preferred_dom), end)
        yield start, next_start
        start = next_start


def iter_quarters(start, end):
    """
    Generate a sequence of tuples representing the span of a quarter
    (quarter_start, quarter_end) where starts are inclusive and ends are
    exclusive (the end of one quarter is the same as the start of the next
    quarter).
    """
    months = iter_months(start, end)
    for quarter_start, quarter_end in months:
        for m in range(2):
            next_month = next(months, None)
            if next_month:
                quarter_end = next_month[1]
        yield quarter_start, quarter_end


def iter_years(start, end):
    """
    Generate a sequence of tuples representing the span of a year (year_start,
    year_end) where starts are inclusive and ends are exclusive (the end of one
    year is the same as the start of the next year).
    """
    months = iter_months(start, end)
    for year_start, year_end in months:
        for m in range(11):
            next_month = next(months, None)
            if next_month:
                year_end = next_month[1]
        yield year_start, year_end


def next_month(dt, preferred_dom=None, delta=1):
    """
    Return dt adjusted to the next month, like dt + timedelta(months=delta) if
    timedelta supported that. If the current day (or preferred_dom) doesn't
    exist in the next month then the day will be adjusted to the closest
    available (for example Jan 31 -> Feb 28).
    """
    if delta == 0:
        raise ValueError("delta must be non-zero")

    # adjust month forward/backward according to delta
    next_year, next_month = dt.year, dt.month
    while delta < 0:
        next_year -= (next_month == 1)
        next_month = next_month - 1 or 12
        delta += 1
    while delta > 0:
        next_year += (next_month == 12)
        next_month = next_month % 12 + 1
        delta -= 1

    return dt.replace(
        year=next_year,
        month=next_month,
        day=nearest_dom(next_year, next_month, preferred_dom or dt.day)
    )


def nearest_dom(year, month, day):
    """
    Return day adjusted as necessary to fit within the days available in
    year/month. For example:

        nearest_dom(2017, 2, 30)  #=> 28

    """
    return min(calendar.monthrange(year, month)[1], day)



from __future__ import absolute_import, unicode_literals

import argparse
import dateutil.parser
from django.core.management.base import CommandError
from django.utils import timezone


class DateAction(argparse.Action):
    """
    argparse action that parses its argument into a timezone-aware datetime
    """
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            dt = dateutil.parser.parse(values)
        except ValueError:
            raise CommandError("can't parse date: {}".format(values))
        if dt.tzinfo is None:
            dt = timezone.make_aware(dt)
        setattr(namespace, self.dest, dt)



from __future__ import absolute_import, unicode_literals

import csv
from tabulate import tabulate
from ._commands import ReportCommand


class Command(ReportCommand):
    help = "Shows registered, activated and active users"

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--csv', action='store_true',
                            help="output csv instead of table")

    def handle(self, *args, **options):
        headers, rows = super(Command, self).handle(*args, **options)

        if options['csv']:
            writer = csv.writer(self.stdout)
            writer.writerow(headers)
            writer.writerows(rows)
        else:
            return tabulate(rows, headers)  # prints on stdout



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Souvenir',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('when', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-when'],
            },
        ),
    ]



from __future__ import absolute_import, unicode_literals

from django.contrib.auth import get_user_model
from django.utils import timezone
import factory
from factory.django import DjangoModelFactory
from souvenirs.models import Souvenir


class UserFactory(DjangoModelFactory):
    class Meta:
        model = get_user_model()

    username = factory.Sequence(lambda n: 'user{}'.format(n))
    password = factory.PostGenerationMethodCall('set_password', 'password')
    is_active = True


class SouvenirFactory(DjangoModelFactory):
    class Meta:
        model = Souvenir

    user = factory.SubFactory(
        UserFactory,
        date_joined=factory.SelfAttribute('..when'),
        last_login=factory.SelfAttribute('..when'),
    )



from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Souvenir(models.Model):
    """
    One instance of seeing an active user
    """
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    when = models.DateTimeField(default=timezone.now, db_index=True)

    class Meta:
        ordering = ['-when']

    def __str__(self):
        return 'user={} when={}'.format(self.user_id, self.when)



from __future__ import absolute_import, unicode_literals

from datetime import datetime
from django.utils import timezone
import pytest
from .factories import SouvenirFactory
from souvenirs.reports import (daily_usage,
                               customer_monthly_usage,
                               customer_quarterly_usage,
                               customer_yearly_usage,
                               calendar_monthly_usage,
                               usage_for_periods,
                               registered_users_as_of)


@pytest.mark.django_db
class TestSouvenir:

    @pytest.fixture(autouse=True)
    def setup(self, db, mocker):
        self.tzinfo = timezone.get_current_timezone()
        self.subscription_start = datetime(
            year=2010, month=1, day=24, hour=22, tzinfo=self.tzinfo)
        self.souvenirs = [
            SouvenirFactory(when=datetime(
                year=i, month=2, day=14, hour=12, tzinfo=self.tzinfo))
            for i in range(2010, 2018)
        ]
        self.souvenirs.append(
            SouvenirFactory(when=datetime(year=2015, month=10, day=17,
                                          hour=12, tzinfo=self.tzinfo))
        )
        mocked_now = mocker.patch('souvenirs.reports.timezone.now')
        mocked_now.return_value = self.now = datetime(
            year=2017, month=4, day=3, hour=23, tzinfo=self.tzinfo)

    def test_daily_usage(self):
        daus = list(daily_usage(self.subscription_start))
        assert len(daus) == 2627  # 7 years, 70 days, 2 leap days

        make_when = lambda y, m, **kw: datetime(
            year=2010, month=1, day=24, hour=22, tzinfo=self.tzinfo).replace(
                year=y, month=m, **kw)

        make_usage = lambda yn, mn, qn, usage, start, end: {
            'period': {
                'start': start,
                'end': end,
            },
            'labels': {
                'year_month': 'Y{:02d} M{:02d}'.format(yn, mn),
                'year_quarter': 'Y{:02d} Q{}'.format(yn, qn),
                'year': 'Y{:02d}'.format(yn),
            },
            'usage': {
                'registered_users': usage[0],
                'activated_users': usage[1],
                'active_users': usage[2],
            },
        }

        non_zero_daus = [dau for dau in daus if dau['usage']['active_users']]
        assert non_zero_daus == [
            make_usage(yn=1, mn=1, qn=1, usage=(1, 1, 1), start=make_when(2010, 2, day=13), end=make_when(2010, 2, day=14)),
            make_usage(yn=2, mn=1, qn=1, usage=(2, 2, 1), start=make_when(2011, 2, day=13), end=make_when(2011, 2, day=14)),
            make_usage(yn=3, mn=1, qn=1, usage=(3, 3, 1), start=make_when(2012, 2, day=13), end=make_when(2012, 2, day=14)),
            make_usage(yn=4, mn=1, qn=1, usage=(4, 4, 1), start=make_when(2013, 2, day=13), end=make_when(2013, 2, day=14)),
            make_usage(yn=5, mn=1, qn=1, usage=(5, 5, 1), start=make_when(2014, 2, day=13), end=make_when(2014, 2, day=14)),
            make_usage(yn=6, mn=1, qn=1, usage=(6, 6, 1), start=make_when(2015, 2, day=13), end=make_when(2015, 2, day=14)),
            make_usage(yn=6, mn=9, qn=3, usage=(7, 7, 1), start=make_when(2015, 10, day=16), end=make_when(2015, 10, day=17)),
            make_usage(yn=7, mn=1, qn=1, usage=(8, 8, 1), start=make_when(2016, 2, day=13), end=make_when(2016, 2, day=14)),
            make_usage(yn=8, mn=1, qn=1, usage=(9, 9, 1), start=make_when(2017, 2, day=13), end=make_when(2017, 2, day=14)),
        ]

        # test number of months returned with custom start/end dates
        when = make_when(2013, 2)
        assert len(list(daily_usage(self.subscription_start, start=when))) == 1500  # 4 years, 39 days, 1 leap day
        assert len(list(daily_usage(self.subscription_start, end=when))) == 1127  # 3 years, 31 days, 1 leap day
        assert len(list(daily_usage(self.subscription_start, start=when, end=when))) == 0

    def test_customer_monthly_usage(self):
        maus = list(customer_monthly_usage(self.subscription_start))
        assert len(maus) == 87  # 7 years, 3 months

        make_when = lambda y, m, **kw: datetime(
            year=2010, month=1, day=24, hour=22, tzinfo=self.tzinfo).replace(
                year=y, month=m, **kw)

        make_usage = lambda yn, mn, qn, usage, start, end: {
            'period': {
                'start': start,
                'end': end,
            },
            'labels': {
                'year_month': 'Y{:02d} M{:02d}'.format(yn, mn),
                'year_quarter': 'Y{:02d} Q{}'.format(yn, qn),
                'year': 'Y{:02d}'.format(yn),
            },
            'usage': {
                'registered_users': usage[0],
                'activated_users': usage[1],
                'active_users': usage[2],
            },
        }

        non_zero_maus = [mau for mau in maus if mau['usage']['active_users']]
        assert non_zero_maus == [
            make_usage(yn=1, mn=1, qn=1, usage=(1, 1, 1), start=make_when(2010, 1), end=make_when(2010, 2)),
            make_usage(yn=2, mn=1, qn=1, usage=(2, 2, 1), start=make_when(2011, 1), end=make_when(2011, 2)),
            make_usage(yn=3, mn=1, qn=1, usage=(3, 3, 1), start=make_when(2012, 1), end=make_when(2012, 2)),
            make_usage(yn=4, mn=1, qn=1, usage=(4, 4, 1), start=make_when(2013, 1), end=make_when(2013, 2)),
            make_usage(yn=5, mn=1, qn=1, usage=(5, 5, 1), start=make_when(2014, 1), end=make_when(2014, 2)),
            make_usage(yn=6, mn=1, qn=1, usage=(6, 6, 1), start=make_when(2015, 1), end=make_when(2015, 2)),
            make_usage(yn=6, mn=9, qn=3, usage=(7, 7, 1), start=make_when(2015, 9), end=make_when(2015, 10)),
            make_usage(yn=7, mn=1, qn=1, usage=(8, 8, 1), start=make_when(2016, 1), end=make_when(2016, 2)),
            make_usage(yn=8, mn=1, qn=1, usage=(9, 9, 1), start=make_when(2017, 1), end=make_when(2017, 2)),
        ]

        # test number of months returned with custom start/end dates
        when = make_when(2013, 2)
        assert len(list(customer_monthly_usage(self.subscription_start, start=when))) == 50  # 4 years, 2 months
        assert len(list(customer_monthly_usage(self.subscription_start, end=when))) == 37  # 3 years, 1 month
        assert len(list(customer_monthly_usage(self.subscription_start, start=when, end=when))) == 0

    def test_customer_quarterly_usage(self):
        qaus = list(customer_quarterly_usage(self.subscription_start))
        assert len(qaus) == 29  # 7 years, 3 months

        make_when = lambda y, m, **kw: datetime(
            year=2010, month=1, day=24, hour=22, tzinfo=self.tzinfo).replace(
                year=y, month=m, **kw)

        make_usage = lambda yn, qn, usage, start, end: {
            'period': {
                'start': start,
                'end': end,
            },
            'labels': {
                'year_quarter': 'Y{:02d} Q{}'.format(yn, qn),
                'year': 'Y{:02d}'.format(yn),
            },
            'usage': {
                'registered_users': usage[0],
                'activated_users': usage[1],
                'active_users': usage[2],
            },
        }

        non_zero_qaus = [qau for qau in qaus if qau['usage']['active_users']]
        assert non_zero_qaus == [
            make_usage(yn=1, qn=1, usage=(1, 1, 1), start=make_when(2010, 1), end=make_when(2010, 4)),
            make_usage(yn=2, qn=1, usage=(2, 2, 1), start=make_when(2011, 1), end=make_when(2011, 4)),
            make_usage(yn=3, qn=1, usage=(3, 3, 1), start=make_when(2012, 1), end=make_when(2012, 4)),
            make_usage(yn=4, qn=1, usage=(4, 4, 1), start=make_when(2013, 1), end=make_when(2013, 4)),
            make_usage(yn=5, qn=1, usage=(5, 5, 1), start=make_when(2014, 1), end=make_when(2014, 4)),
            make_usage(yn=6, qn=1, usage=(6, 6, 1), start=make_when(2015, 1), end=make_when(2015, 4)),
            make_usage(yn=6, qn=3, usage=(7, 7, 1), start=make_when(2015, 7), end=make_when(2015, 10)),
            make_usage(yn=7, qn=1, usage=(8, 8, 1), start=make_when(2016, 1), end=make_when(2016, 4)),
            make_usage(yn=8, qn=1, usage=(9, 9, 1), start=make_when(2017, 1), end=self.now),
        ]

        # test number of quarters returned with custom start/end dates
        when = make_when(2013, 2)
        assert len(list(customer_quarterly_usage(self.subscription_start, start=when))) == 17  # 4 years, 2 months
        assert len(list(customer_quarterly_usage(self.subscription_start, end=when))) == 13  # 3 years, 1 month
        assert len(list(customer_quarterly_usage(self.subscription_start, start=when, end=when))) == 0

    def test_customer_yearly_usage(self):
        yaus = list(customer_yearly_usage(self.subscription_start))
        assert len(yaus) == 8  # 7 years, 3 months

        make_when = lambda y, m, **kw: datetime(
            year=2010, month=1, day=24, hour=22, tzinfo=self.tzinfo).replace(
                year=y, month=m, **kw)

        make_usage = lambda yn, usage, start, end: {
            'period': {
                'start': start,
                'end': end,
            },
            'labels': {
                'year': 'Y{:02d}'.format(yn),
            },
            'usage': {
                'registered_users': usage[0],
                'activated_users': usage[1],
                'active_users': usage[2],
            },
        }

        non_zero_yaus = [yau for yau in yaus if yau['usage']['active_users']]
        assert non_zero_yaus == [
            make_usage(yn=1, usage=(1, 1, 1), start=make_when(2010, 1), end=make_when(2011, 1)),
            make_usage(yn=2, usage=(2, 2, 1), start=make_when(2011, 1), end=make_when(2012, 1)),
            make_usage(yn=3, usage=(3, 3, 1), start=make_when(2012, 1), end=make_when(2013, 1)),
            make_usage(yn=4, usage=(4, 4, 1), start=make_when(2013, 1), end=make_when(2014, 1)),
            make_usage(yn=5, usage=(5, 5, 1), start=make_when(2014, 1), end=make_when(2015, 1)),
            make_usage(yn=6, usage=(7, 7, 2), start=make_when(2015, 1), end=make_when(2016, 1)),
            make_usage(yn=7, usage=(8, 8, 1), start=make_when(2016, 1), end=make_when(2017, 1)),
            make_usage(yn=8, usage=(9, 9, 1), start=make_when(2017, 1), end=self.now),
        ]

        # test number of quarters returned with custom start/end dates
        when = make_when(2013, 2)
        assert len(list(customer_yearly_usage(self.subscription_start, start=when))) == 5  # 4 years, 2 months
        assert len(list(customer_yearly_usage(self.subscription_start, end=when))) == 4  # 3 years, 1 month
        assert len(list(customer_yearly_usage(self.subscription_start, start=when, end=when))) == 0

    def test_calendar_monthly_usage(self):
        cal_maus = list(calendar_monthly_usage(self.subscription_start))
        assert len(cal_maus) == 88  # 7 years, 4 months

        make_when = lambda y, m, **kw: datetime(
            year=2010, month=1, day=1, hour=0, tzinfo=self.tzinfo).replace(
                year=y, month=m, **kw)

        make_usage = lambda usage, start, end: {
            'period': {
             'start': start,
             'end': end,
            },
            'labels': {
             'calendar_year_month': start.strftime('%Y-%m'),
             'calendar_year': start.year,
            },
            'usage': {
             'registered_users': usage[0],
             'activated_users': usage[1],
             'active_users': usage[2],
            },
        }

        non_zero_cal_maus = [mau for mau in cal_maus if mau['usage']['active_users']]
        assert non_zero_cal_maus == [
            make_usage(usage=(1, 1, 1), start=make_when(2010, 2), end=make_when(2010, 3)),
            make_usage(usage=(2, 2, 1), start=make_when(2011, 2), end=make_when(2011, 3)),
            make_usage(usage=(3, 3, 1), start=make_when(2012, 2), end=make_when(2012, 3)),
            make_usage(usage=(4, 4, 1), start=make_when(2013, 2), end=make_when(2013, 3)),
            make_usage(usage=(5, 5, 1), start=make_when(2014, 2), end=make_when(2014, 3)),
            make_usage(usage=(6, 6, 1), start=make_when(2015, 2), end=make_when(2015, 3)),
            make_usage(usage=(7, 7, 1), start=make_when(2015, 10), end=make_when(2015, 11)),
            make_usage(usage=(8, 8, 1), start=make_when(2016, 2), end=make_when(2016, 3)),
            make_usage(usage=(9, 9, 1), start=make_when(2017, 2), end=make_when(2017, 3)),
        ]

        # test number of months returned with custom start/end dates
        assert len(list(calendar_monthly_usage(start=make_when(2013, 2), end=make_when(2013,5)))) == 3
        assert len(list(calendar_monthly_usage(start=make_when(2013, 2), end=make_when(2013,2)))) == 0



from __future__ import absolute_import, unicode_literals

import datetime
from django.utils import timezone
import pytest
from souvenirs.models import Souvenir
from souvenirs.control import count_active_users, souvenez
from .factories import SouvenirFactory, UserFactory


@pytest.mark.django_db
class TestSouvenir:

    @pytest.fixture(autouse=True)
    def setup(self, db):
        self.tzinfo = timezone.get_current_timezone()
        self.souvenirs = [
            SouvenirFactory(when=datetime.datetime(year=i, month=2, day=14,
                                                   hour=12, tzinfo=self.tzinfo))
            for i in range(2010, 2018)
        ]

    def test_souvenez(self):
        Souvenir.objects.all().delete()
        assert Souvenir.objects.count() == 0
        u, u2 = UserFactory(), UserFactory()
        assert souvenez(u, when=timezone.now() - datetime.timedelta(minutes=61)) == 'added'
        assert souvenez(u) == 'added'
        assert souvenez(u2) == 'added'
        assert souvenez(u) == 'rate-limited'
        assert souvenez(u, ratelimit=False) == 'added'
        assert Souvenir.objects.count() == 4

    def test_count_active_users(self):
        assert count_active_users() == 8

        when = datetime.datetime(year=2013, month=1, day=1, tzinfo=self.tzinfo)
        assert count_active_users(start=when)
        assert count_active_users(end=when)

        # combine start and end.
        when = datetime.datetime(year=2013, month=2, day=1, tzinfo=self.tzinfo)
        assert count_active_users(start=when, end=when.replace(year=2015)) == 2
        assert count_active_users(start=when.replace(year=2015), end=when) == 0

        # ensure start inclusive and end exclusive by using a datetime that
        # exactly matches a record.
        when = datetime.datetime(year=2013, month=2, day=14, hour=12, tzinfo=self.tzinfo)
        assert Souvenir.objects.filter(when=when).exists()
        assert count_active_users(start=when) == 5
        assert count_active_users(end=when) == 3
        assert count_active_users(start=when, end=when) == 0

    def test_count_active_users_distinct(self):
        for s in self.souvenirs:
            SouvenirFactory(user=s.user, when=s.when)
        self.test_count_active_users()




from __future__ import absolute_import, unicode_literals

from datetime import timedelta
import logging
from django.conf import settings
from django.core.cache import caches
from django.utils import timezone
from .models import Souvenir


logger = logging.getLogger(__name__)


def souvenez(user, when=None, ratelimit=True, check_duplicate=False):
    """
    Save a Souvenir to the DB, rate-limited by default to once per hour.
    Returns a string: "added", "rate-limited" or "duplicated".
    """
    # user can be a User object or PK (for backfill script)
    user_id = getattr(user, 'id', user)
    username = getattr(user, 'username', user)  # just for logging

    if when is None:
        when = timezone.now()

    if ratelimit is True:
        ratelimit = getattr(settings, 'SOUVENIRS_RATELIMIT_SECONDS', 360)

    if ratelimit:
        name = getattr(settings, 'SOUVENIRS_CACHE_NAME', 'default')
        prefix = getattr(settings, 'SOUVENIRS_CACHE_PREFIX', 'souvenir.')
        key = '{}.{}'.format(prefix, user_id)
        cache = caches[name]
        value = cache.get(key)
        if value and when < value + timedelta(seconds=ratelimit):
            logger.debug("rate-limited %s (last seen %s)".format(username, value))
            return 'rate-limited'
        cache.set(key, when)

    if check_duplicate:
        if Souvenir.objects.filter(user_id=user_id, when=when).exists():
            logger.debug("ignoring duplicate souvenir for %s (%s)".format(username, when))
            return 'duplicated'

    Souvenir(user_id=user_id, when=when).save()
    logger.debug("saved souvenir for %s (%s)".format(username, when))
    return 'added'


def count_active_users(start=None, end=None, qs=None):
    """
    Return the number of active users between start and end datetimes,
    inclusive and exclusive respectively.
    """
    if qs is None:
        qs = Souvenir.objects.all()
    if start:
        qs = qs.filter(when__gte=start)  # inclusive
    if end:
        qs = qs.filter(when__lt=end)     # exclusive
    return qs.values('user').distinct().count()



from __future__ import absolute_import, unicode_literals

from django.core.management.base import CommandError
from django.core.management.base import BaseCommand, CommandError
from souvenirs.reports import (daily_usage, customer_monthly_usage,
                               customer_quarterly_usage, customer_yearly_usage)
from ._helpers import DateAction


class ReportCommand(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--subscription-start', metavar='DATE', action=DateAction,
                            help="customer subscription start or renewal date")
        parser.add_argument('--after', metavar='DATE', action=DateAction,
                            help="include lines starting at (inclusive)")
        parser.add_argument('--before', metavar='DATE', action=DateAction,
                            help="include lines ending at (exclusive)")

        parser.add_argument('--daily', action='store_const', dest='report', const='daily',
                            help="report on daily activity")
        parser.add_argument('--monthly', action='store_const', dest='report', const='monthly',
                            help="report on monthly activity (default)")
        parser.add_argument('--quarterly', action='store_const', dest='report', const='quarterly',
                            help="report on quarterly activity")
        parser.add_argument('--yearly', action='store_const', dest='report', const='yearly',
                            help="report on yearly activity")

        parser.add_argument('--recent', nargs='?', const=1, type=int, metavar='NUM',
                            help="show only the most recent N (1) entries")
        parser.add_argument('--ascending', action='store_true',
                            help="output in date-ascending order (default: descending)")
        parser.add_argument('--datefmt', default='%Y-%m-%d',
                            help="strftime for date columns (default: %Y-%m-%d)")

    def handle(self, *args, **options):
        report = options['report'] or 'monthly'

        if (report in ['daily', 'monthly', 'quarterly', 'yearly'] and
            options['subscription_start'] is None
        ):
            raise CommandError("{} report requires --subscription-start"
                               .format(report))

        report_method = getattr(self, '{}_report'.format(report))
        headers, rows = report_method(options)

        if options['recent']:
            rows = rows[-options['recent']:]

        # reports are chronologically ascending by default (mainly because of
        # enumerations), but for display we prefer reversed by default.
        if not options['ascending']:
            rows = reversed(rows)

        return headers, rows

    def daily_report(self, options):
        headers = ['date', 'registered', 'activated', 'active']
        usage = daily_usage(
            subscription_start=options['subscription_start'],
            start=options['after'],
            end=options['before'],
        )
        rows = [
            [d['period']['end'].strftime(options['datefmt']),
             d['usage']['registered_users'],
             d['usage']['activated_users'],
             d['usage']['active_users'],
            ] for d in usage
        ]
        return headers, rows

    def monthly_report(self, options):
        headers = ['month', 'start', 'end', 'registered', 'activated', 'active']
        usage = customer_monthly_usage(
            subscription_start=options['subscription_start'],
            start=options['after'],
            end=options['before'],
        )
        rows = [
            [d['labels']['year_month'],
             d['period']['start'].strftime(options['datefmt']),
             d['period']['end'].strftime(options['datefmt']),
             d['usage']['registered_users'],
             d['usage']['activated_users'],
             d['usage']['active_users'],
            ] for d in usage
        ]
        return headers, rows

    def quarterly_report(self, options):
        headers = ['quarter', 'start', 'end', 'registered', 'activated', 'active']
        usage = customer_quarterly_usage(
            subscription_start=options['subscription_start'],
            start=options['after'],
            end=options['before'],
        )
        rows = [
            [d['labels']['year_quarter'],
             d['period']['start'].strftime(options['datefmt']),
             d['period']['end'].strftime(options['datefmt']),
             d['usage']['registered_users'],
             d['usage']['activated_users'],
             d['usage']['active_users'],
            ] for d in usage
        ]
        return headers, rows

    def yearly_report(self, options):
        headers = ['year', 'start', 'end', 'registered', 'activated', 'active']
        usage = customer_yearly_usage(
            subscription_start=options['subscription_start'],
            start=options['after'],
            end=options['before'],
        )
        rows = [
            [d['labels']['year'],
             d['period']['start'].strftime(options['datefmt']),
             d['period']['end'].strftime(options['datefmt']),
             d['usage']['registered_users'],
             d['usage']['activated_users'],
             d['usage']['active_users'],
            ] for d in usage
        ]
        return headers, rows



from __future__ import absolute_import, unicode_literals

import itertools
from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils import timezone
from django.utils.module_loading import import_string
from .control import count_active_users
from .utils import (iter_days, iter_quarters, iter_months, iter_years,
                    adjust_to_calendar_month)


izip = getattr(itertools, 'izip', zip)


def daily_usage(subscription_start, start=None, end=None):
    # labeling depends on having a month number, so defer to
    # customer_monthly_usage for that.
    for monthly_usage in customer_monthly_usage(subscription_start,
                                                start=start,
                                                end=end):
        periods = iter_days(monthly_usage['period']['start'],
                            monthly_usage['period']['end'])
        for daily_usage in usage_for_periods(periods):
            usage = dict(daily_usage,
                         labels=monthly_usage['labels'])
            yield usage


def customer_monthly_usage(subscription_start, start=None, end=None):
    if start is None:
        start = subscription_start

    # regardless of start, the monthly iterator must use subscription_start for
    # the sake of enumerating.
    periods = iter_months(start=subscription_start,
                          end=end or timezone.now())

    for m, usage in enumerate(usage_for_periods(periods), 1):
        if usage['period']['end'] <= start:
            continue
        usage.update(
            labels=dict(
                year_month=label_year_month_m(m),
                year_quarter=label_year_quarter_m(m),
                year=label_year_m(m),
            ),
        )
        yield usage


def customer_quarterly_usage(subscription_start, start=None, end=None):
    if start is None:
        start = subscription_start

    # regardless of start, the quarterly iterator must use subscription_start
    # for the sake of enumerating.
    periods = iter_quarters(start=subscription_start,
                            end=end or timezone.now())

    for q, usage in enumerate(usage_for_periods(periods), 1):
        if usage['period']['end'] <= start:
            continue
        usage.update(
            labels=dict(
                year_quarter=label_year_quarter_q(q),
                year=label_year_q(q),
            ),
        )
        yield usage


def customer_yearly_usage(subscription_start, start=None, end=None):
    if start is None:
        start = subscription_start

    # regardless of start, the yearly iterator must use subscription_start
    # for the sake of enumerating.
    periods = iter_years(start=subscription_start,
                         end=end or timezone.now())

    for y, usage in enumerate(usage_for_periods(periods), 1):
        if usage['period']['end'] <= start:
            continue
        usage.update(
            labels=dict(
                year=label_year_y(y)
            ),
        )
        yield usage


def calendar_monthly_usage(start, end=None):
    start = adjust_to_calendar_month(start)
    periods = iter_months(start, end or timezone.now())
    for usage in usage_for_periods(periods):
        usage.update(
            labels=dict(
                calendar_year_month=label_calendar_year_month(usage['period']['start']),
                calendar_year=label_calendar_year(usage['period']['start']),
            ),
        )
        yield usage


month_to_year = lambda m: (m - 1) // 12 + 1
month_to_month = lambda m: (m - 1) % 12 + 1
month_to_quarter = lambda m: ((m - 1) // 3) % 4 + 1
quarter_to_year = lambda q: (q - 1) // 4 + 1
quarter_to_quarter = lambda q: (q - 1) % 4 + 1
label_year_month_m = lambda m: 'Y{:02d} M{:02d}'.format(month_to_year(m), month_to_month(m))
label_year_quarter_q = lambda q: 'Y{:02d} Q{}'.format(quarter_to_year(q), quarter_to_quarter(q))
label_year_quarter_m = lambda m: 'Y{:02d} Q{}'.format(month_to_year(m), month_to_quarter(m))
label_year_y = lambda y: 'Y{:02d}'.format(y)
label_year_q = lambda q: 'Y{:02d}'.format(quarter_to_year(q))
label_year_m = lambda m: 'Y{:02d}'.format(month_to_year(m))
label_calendar_year_month = lambda d: d.strftime('%Y-%m')
label_calendar_year = lambda d: d.year  # int


def usage_for_periods(*args, **kwargs):
    name = getattr(settings, 'SOUVENIRS_USAGE_REPORTS_FUNCTION', None)
    func = import_string(name) if name else _usage_for_periods
    return func(*args, **kwargs)


def _usage_for_periods(periods):
    """
    Generate a sequence of dictionaries of usage data corresponding to periods,
    each of which should be a tuple of (start, end) datetimes, where start is
    inclusive and end is exclusive.

    Each dictionary in the generated sequence has this form:

        {
            period: {
                start: datetime,
                end: datetime,
            }
            usage: {
                registered_users: int,
                activated_users: int,
                active_users: int,
            }
        }

    """
    rp, ap, periods = itertools.tee(periods, 3)
    ir = (registered_users_as_of(end) for start, end in rp)
    ia = (count_active_users(*p) for p in ap)
    for p, r, active in izip(periods, ir, ia):
        start, end = p
        registered, activated = r
        yield dict(
            period=dict(
                start=start,
                end=end,
            ),
            usage=dict(
                registered_users=registered,
                activated_users=activated,
                active_users=active,
            ),
        )


def registered_users_as_of(date):
    """
    Return a tuple of the form (registered, activated) indicating the total
    registered and activated users for the given date.
    """
    User = get_user_model()
    users = User.objects.filter(date_joined__lt=date)
    return users.count(), users.filter(is_active=True).count()



from __future__ import absolute_import, unicode_literals

from codecs import open
import os
from setuptools import setup, find_packages
import souvenirs as app


def read(fname):
    top = os.path.dirname(__file__)
    with open(os.path.join(top, fname), encoding='utf-8') as f:
        return f.read()


setup(
    name="django-souvenirs",
    version=app.__version__,
    description='Django app for efficiently measuring usage',
    long_description=read('README.rst'),
    license='MIT',
    platforms=['OS Independent'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='django reusable souvenirs metrics usage tracking'.split(),
    author='Aron Griffis',
    author_email='aron@scampersand.com',
    url="https://github.com/appsembler/django-souvenirs",
    packages=find_packages(),
    install_requires=['tabulate'],
)



from __future__ import absolute_import, unicode_literals

from datetime import timedelta
from django.contrib.auth.models import AnonymousUser
from django.utils import timezone
import pytest
from souvenirs.middleware import SouvenirsMiddleware
from souvenirs.models import Souvenir
from .factories import UserFactory


@pytest.mark.django_db
class TestMiddleware:

    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        self.sm = SouvenirsMiddleware()
        self.request = mocker.Mock()

    def test_process_request_with_user(self):
        self.request.user = u = UserFactory()
        assert Souvenir.objects.count() == 0
        assert self.sm.process_request(self.request) is None
        assert Souvenir.objects.count() == 1
        s = Souvenir.objects.all()[0]
        assert s.user == u
        assert timedelta() <= (timezone.now() - s.when) <= timedelta(seconds=10)

    def test_process_request_with_anonymous(self):
        self.request.user = u = AnonymousUser()
        assert Souvenir.objects.count() == 0
        assert self.sm.process_request(self.request) is None
        assert Souvenir.objects.count() == 0
